from functools import partial
from typing import Optional

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.coupler import coupler
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.routing.get_route import get_route
from gdsfactory.types import ComponentSpec, CrossSectionSpec


@cell
def mzi(
    delta_length: float = 10.0,
    length_y: float = 2.0,
    length_x: Optional[float] = 0.1,
    bend: ComponentSpec = "bend_euler",
    straight: ComponentSpec = straight_function,
    straight_y: Optional[ComponentSpec] = None,
    straight_x_top: Optional[ComponentSpec] = None,
    straight_x_bot: Optional[ComponentSpec] = None,
    splitter: ComponentSpec = "mmi1x2",
    combiner: Optional[ComponentSpec] = None,
    with_splitter: bool = True,
    port_e1_splitter: str = "o2",
    port_e0_splitter: str = "o3",
    port_e1_combiner: str = "o2",
    port_e0_combiner: str = "o3",
    nbends: int = 2,
    cross_section: CrossSectionSpec = "strip",
    cross_section_x_top: Optional[CrossSectionSpec] = None,
    cross_section_x_bot: Optional[CrossSectionSpec] = None,
) -> Component:
    """Mzi.

    Args:
        delta_length: bottom arm vertical extra length.
        length_y: vertical length for both and top arms.
        length_x: horizontal length. None uses to the straight_x_bot/top defaults.
        bend: 90 degrees bend library.
        straight: straight function.
        straight_y: straight for length_y and delta_length.
        straight_x_top: top straight for length_x.
        straight_x_bot: bottom straight for length_x.
        splitter: splitter function.
        combiner: combiner function.
        with_splitter: if False removes splitter.
        port_e1_combiner: east top combiner port.
        port_e0_splitter: east bot splitter port.
        port_e1_splitter: east top splitter port.
        port_e0_combiner: east bot combiner port.
        nbends: from straight top/bot to combiner (at least 2).
        cross_section: for routing (sxtop/sxbot to combiner).
        cross_section_x_top: optional top cross_section (defaults to cross_section).
        cross_section_x_bot: optional bottom cross_section (defaults to cross_section).

    .. code::

                       b2______b3
                      |  sxtop  |
              straight_y        |
                      |         |
                      b1        b4
            splitter==|         |==combiner
                      b5        b8
                      |         |
              straight_y        |
                      |         |
        delta_length/2          |
                      |         |
                     b6__sxbot__b7
                          Lx
    """
    combiner = combiner or splitter

    straight_x_top = straight_x_top or straight
    straight_x_bot = straight_x_bot or straight
    straight_y = straight_y or straight

    cross_section_x_bot = cross_section_x_bot or cross_section
    cross_section_x_top = cross_section_x_top or cross_section

    bend_spec = bend
    bend = gf.get_component(bend, cross_section=cross_section)

    c = Component()
    cp1 = gf.get_component(splitter)
    cp2 = gf.get_component(combiner) if combiner else cp1

    if with_splitter:
        cp1 = c << cp1

    cp2 = c << cp2
    b5 = c << bend
    b5.mirror()
    b5.connect("o1", cp1.ports[port_e0_splitter])

    syl = c << gf.get_component(
        straight_y, length=delta_length / 2 + length_y, cross_section=cross_section
    )
    syl.connect("o1", b5.ports["o2"])
    b6 = c << bend
    b6.connect("o1", syl.ports["o2"])

    straight_x_bot = (
        gf.get_component(
            straight_x_bot, length=length_x, cross_section=cross_section_x_bot
        )
        if length_x
        else gf.get_component(straight_x_bot)
    )
    sxb = c << straight_x_bot
    sxb.connect("o1", b6.ports["o2"])

    b1 = c << bend
    b1.connect("o1", cp1.ports[port_e1_splitter])

    sy = c << gf.get_component(straight_y, length=length_y, cross_section=cross_section)
    sy.connect("o1", b1.ports["o2"])

    b2 = c << bend
    b2.connect("o2", sy.ports["o2"])
    straight_x_top = (
        gf.get_component(
            straight_x_top, length=length_x, cross_section=cross_section_x_top
        )
        if length_x
        else gf.get_component(straight_x_top)
    )
    sxt = c << straight_x_top
    sxt.connect("o1", b2.ports["o1"])

    cp2.mirror()
    cp2.xmin = sxt.ports["o2"].x + bend.info["radius"] * nbends + 0.1

    route = get_route(
        sxt.ports["o2"],
        cp2.ports[port_e1_combiner],
        straight=straight,
        bend=bend_spec,
        cross_section=cross_section,
    )
    c.add(route.references)
    route = get_route(
        sxb.ports["o2"],
        cp2.ports[port_e0_combiner],
        straight=straight,
        bend=bend_spec,
        cross_section=cross_section,
    )
    c.add(route.references)

    if with_splitter:
        c.add_ports(cp1.get_ports_list(orientation=180), prefix="in")
    else:
        c.add_port("o1", port=b1.ports["o1"])
        c.add_port("o2", port=b5.ports["o1"])
    c.add_ports(cp2.get_ports_list(orientation=0), prefix="out")
    c.add_ports(sxt.get_ports_list(port_type="electrical"), prefix="top")
    c.add_ports(sxb.get_ports_list(port_type="electrical"), prefix="bot")
    c.auto_rename_ports()
    return c


mzi1x2 = partial(mzi, splitter=mmi1x2, combiner=mmi1x2)
mzi2x2_2x2 = partial(
    mzi,
    splitter=mmi2x2,
    combiner=mmi2x2,
    port_e1_splitter="o3",
    port_e0_splitter="o4",
    port_e1_combiner="o3",
    port_e0_combiner="o4",
)

mzi1x2_2x2 = partial(
    mzi,
    combiner=mmi2x2,
    port_e1_combiner="o3",
    port_e0_combiner="o4",
)

mzi_coupler = partial(
    mzi2x2_2x2,
    splitter=coupler,
    combiner=coupler,
)


if __name__ == "__main__":
    # from gdsfactory.cross_section import heater_metal
    # c = mzi(cross_section=heater_metal)
    # from gdsfactory.cross_section import strip_heater_metal
    # c = mzi(cross_section_x_top=strip_heater_metal, length_x=100)

    c = mzi()
    c.show()

    # WIDTH = 2
    # LAYER = (34, 0)
    # xs_metal = gf.partial(strip, width=WIDTH, layer=LAYER)
    # mmi1x2 = gf.partial(
    #     gf.components.mmi1x2,
    #     cross_section=xs_metal,
    #     width=WIDTH,
    #     width_taper=WIDTH,
    #     width_mmi=3 * WIDTH,
    # )
    # mzi = gf.partial(gf.components.mzi, cross_section=xs_metal, splitter=mmi1x2)
    # c = mzi()

    # c = gf.c.mzi2x2_2x2(straight_x_top="straight_heater_metal")
    # c.show()

    # extend_ports2 = gf.partial(gf.components.extend_ports, length=10)

    # straigth_extended2 = gf.compose(
    #     extend_ports2, gf.partial(gf.components.straight, width=0.9)
    # )
    # c = straigth_extended2()
    # c.show()

    # delta_length = 116.8 / 2
    # print(delta_length)
    # c = mzi(delta_length=delta_length, with_splitter=False)
    # c.pprint_netlist()
    # mmi2x2 = gf.partial(gf.components.mmi2x2, width_mmi=5, gap_mmi=2)
    # c = mzi(delta_length=10, combiner=gf.components.mmi1x2, splitter=mmi2x2)

    # c = mzi1x2_2x2()
    # c = mzi_coupler(length_x=5)
    # c = mzi2x2()

    # c = mzi2x2_2x2(straight_x_top="straight_heater_metal")
    # c.show()

    # c = mzi(
    #     delta_length=100,
    #     straight_x_top="straight_pin",
    #     # straight_x_top=gf.components.straight_heater_meander,
    #     # straight_x_bot=gf.components.straight_heater_meander,
    #     # straight_x_top=gf.components.straight_heater_metal,
    #     # straight_x_bot=gf.components.straight_heater_metal,
    #     # length_x=None,
    #     length_x=300,
    #     length_y=200.8,
    #     # with_splitter=False,
    #     splitter="mmi1x2",
    # )
    # c.show(show_ports=False)
    # c.show(show_subports=True)
    # c.pprint()
    # n = c.get_netlist()
    # c.plot()
    # print(c.settings)



"""tidy3d mode solver.

tidy3d has a powerful open source mode solver.

tidy3d can:

- compute bend modes.

TODO:

- fix mode overlaps for bend loss
- calculate dispersion

Maybe:

- combine modes package, mpb and tidy3d APIs
"""

import pathlib
import pickle
from types import SimpleNamespace
from typing import Callable, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from pydantic import BaseModel, Extra
from scipy.constants import c as SPEED_OF_LIGHT
from tidy3d.plugins.mode.solver import compute_modes

from gdsfactory.config import CONFIG, logger
from gdsfactory.serialization import get_hash
from gdsfactory.simulation.gtidy3d.materials import si, sio2
from gdsfactory.types import PathType


def plot(
    X,
    Y,
    n,
    mode=None,
    num_levels: int = 8,
    n_cmap=None,
    mode_cmap=None,
    axes="xy",
    title=None,
    normalize_mode=False,
) -> None:
    """Plot mode in matplotlib"""
    x, y = axes
    if n_cmap is None:
        n_cmap = colors.LinearSegmentedColormap.from_list(
            name="custom", colors=["#ffffff", "#c1d9ed"]
        )
    if mode_cmap is None:
        mode_cmap = "inferno"
    if mode is not None:
        mode = np.abs(mode)
        if normalize_mode:
            mode = mode / mode.max()
        plt.contour(
            X,
            Y,
            mode,
            cmap=mode_cmap,
            levels=np.linspace(mode.min(), mode.max(), num_levels),
        )
        plt.colorbar(label="mode")
    n = np.real(n)
    plt.pcolormesh(X, Y, n, cmap=n_cmap)
    plt.colorbar(label="n")
    plt.xlabel(x)
    plt.ylabel(y)
    plt.grid(True, alpha=0.4)
    plt.axis("scaled")
    if title is not None:
        plt.title(title)


def _mesh_1d(x_min, x_max, mesh_x):
    if isinstance(mesh_x, int):
        x = np.linspace(x_min, x_max, mesh_x + 1)
        dx = x[1] - x[0]
    elif isinstance(mesh_x, float):
        dx = mesh_x
        x = np.arange(x_min, x_max + 0.999 * dx, dx)
        mesh_x = x.shape[0]
    else:
        raise ValueError("Invalid 'mesh_x'")
    return x, mesh_x, dx


def create_mesh(x_min, y_min, x_max, y_max, mesh_x, mesh_y):
    x, mesh_x, dx = _mesh_1d(x_min, x_max, mesh_x)
    y, mesh_y, dy = _mesh_1d(y_min, y_max, mesh_y)

    Yx, Xx = np.meshgrid(y[:-1], x[:-1] + dx / 2)
    Yy, Xy = np.meshgrid(y[:-1] + dy / 2, x[:-1])
    Yz, Xz = np.meshgrid(y[:-1], x[:-1])
    return x, y, Xx, Yx, Xy, Yy, Xz, Yz


def get_n(
    Y,
    Z,
    ncore: float,
    nclad: float,
    wg_width: float,
    t_box: float,
    t_wg: float,
    t_slab: float,
    t_clad: float,
):
    """Return index matrix."""
    w = wg_width
    n = np.ones_like(Y) * nclad
    n[
        (-w / 2 - t_clad / 2 <= Y)
        & (Y <= w / 2 + t_clad / 2)
        & (Z >= t_box)
        & (Z <= t_box + t_wg + t_clad)
    ] = nclad
    n[(Z <= 1.0 + t_slab + t_clad)] = nclad
    n[(-w / 2 <= Y) & (Y <= w / 2) & (Z >= t_box) & (Z <= t_box + t_wg)] = ncore
    n[(Z >= t_box) & (Z <= t_box + t_slab)] = ncore if t_slab else nclad
    return n


SETTINGS = [
    "wavelength",
    "wg_width",
    "t_wg",
    "t_slab",
    "t_box",
    "t_clad",
    "ncore",
    "nclad",
    "w_sim",
    "resolution",
    "nmodes",
    "bend_radius",
]


class Waveguide(BaseModel):
    """Waveguide Model.

    Args:
        wavelength: (um).
        wg_width: waveguide width.
        t_wg: thickness waveguide (um).
        t_slab: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        w_sim: width simulation (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
    """

    wavelength: float = 1.55
    wg_width: float = 0.45
    t_wg: float = 0.22
    t_slab: float = 0.0
    t_box: float = 2.0
    t_clad: float = 2.0
    ncore: Union[float, Callable[[str], float]] = si
    nclad: Union[float, Callable[[str], float]] = sio2
    w_sim: float = 2.0
    resolution: int = 100
    nmodes: int = 4
    bend_radius: Optional[float] = None

    class Config:
        extra = Extra.allow

    @property
    def t_sim(self):
        return self.t_box + self.t_wg + self.t_clad

    def get_ncore(self, wavelength):
        return self.ncore(wavelength) if callable(self.ncore) else self.ncore

    def get_nclad(self, wavelength):
        return self.nclad(wavelength) if callable(self.nclad) else self.nclad

    def plot_index(self, wavelength: Optional[float] = None) -> None:
        wavelength = wavelength or self.wavelength
        x, y, Xx, Yx, Xy, Yy, Xz, Yz = create_mesh(
            -self.w_sim / 2,
            0.0,
            +self.w_sim / 2,
            self.t_sim,
            self.resolution,
            self.resolution,
        )

        nx = get_n(
            Xx,
            Yx,
            wg_width=self.wg_width,
            ncore=self.get_ncore(wavelength),
            nclad=self.get_nclad(wavelength),
            t_box=self.t_box,
            t_slab=self.t_slab,
            t_wg=self.t_wg,
            t_clad=self.t_clad,
        )
        plot(Xx, Yx, nx)
        plt.show()

    def compute_modes(self, wavelength: Optional[float] = None) -> None:
        wavelength = wavelength or self.wavelength
        x, y, Xx, Yx, Xy, Yy, Xz, Yz = create_mesh(
            -self.w_sim / 2,
            0.0,
            +self.w_sim / 2,
            self.t_sim,
            self.resolution,
            self.resolution,
        )

        nx = get_n(
            Xx,
            Yx,
            wg_width=self.wg_width,
            ncore=self.get_ncore(wavelength),
            nclad=self.get_nclad(wavelength),
            t_box=self.t_box,
            t_slab=self.t_slab,
            t_wg=self.t_wg,
            t_clad=self.t_clad,
        )
        ny = get_n(
            Xy,
            Yy,
            wg_width=self.wg_width,
            ncore=self.get_ncore(wavelength),
            nclad=self.get_nclad(wavelength),
            t_box=self.t_box,
            t_slab=self.t_slab,
            t_wg=self.t_wg,
            t_clad=self.t_clad,
        )
        nz = get_n(
            Xz,
            Yz,
            wg_width=self.wg_width,
            ncore=self.get_ncore(wavelength),
            nclad=self.get_nclad(wavelength),
            t_box=self.t_box,
            t_slab=self.t_slab,
            t_wg=self.t_wg,
            t_clad=self.t_clad,
        )

        ((Ex, Ey, Ez), (Hx, Hy, Hz)), neffs = (
            x.squeeze()
            for x in compute_modes(
                eps_cross=[nx**2, ny**2, nz**2],
                coords=[x, y],
                freq=SPEED_OF_LIGHT / (wavelength * 1e-6),
                mode_spec=SimpleNamespace(
                    num_modes=10,
                    bend_radius=self.bend_radius,
                    bend_axis=1,
                    angle_theta=0.0,
                    angle_phi=0.0,
                    num_pml=(0, 0),
                    target_neff=self.get_ncore(wavelength),
                    sort_by="largest_neff",
                ),
            )
        )

        self.nx, self.ny, self.nz = nx, ny, nz
        self.Xx, self.Yx, self.Xy, self.Yy, self.Xz, self.Yz = Xx, Yx, Xy, Yy, Xz, Yz
        self.Ex, self.Ey, self.Ez = Ex, Ey, Ez
        self.Hx, self.Hy, self.Hz = Hx, Hy, Hz
        self.neffs = neffs

    def plot_Ex(self, index: int = 0) -> None:
        if not hasattr(self, "neffs"):
            self.compute_modes()

        nx, neffs, Ex = self.nx, self.neffs, self.Ex
        neff_, Ex_ = np.real(neffs[index]), Ex[..., index]
        plot(self.Xx, self.Yx, nx, mode=np.abs(Ex_) ** 2, title=f"Ex::{neff_:.3f}")
        plt.show()

    def plot_Ey(self, index: int = 0) -> None:
        if not hasattr(self, "neffs"):
            self.compute_modes()

        nx, neffs, Ey = self.nx, self.neffs, self.Ey
        neff_, Ey_ = np.real(neffs[index]), Ey[..., index]
        plot(self.Xx, self.Yx, nx, mode=np.abs(Ey_) ** 2, title=f"Ey::{neff_:.3f}")
        plt.show()

    def _repr_html_(self) -> str:
        """Show index in matplotlib for jupyter notebooks."""
        self.plot_index()
        return self.__repr__()

    def __repr__(self) -> str:
        """Show index in matplotlib for jupyter notebooks."""
        return ", ".join([f"{k}:{getattr(self, k)!r}" for k in SETTINGS])

    def pickle_dump(self, filepath: PathType) -> None:
        data = pickle.dumps(self)
        filepath = pathlib.Path(filepath)
        filepath.write_bytes(data)
        logger.info(f"write {filepath} waveguide to file cache.")


def pickle_load(filepath: PathType) -> Waveguide:
    filepath = pathlib.Path(filepath)
    return pickle.loads(filepath.read_bytes())


def sweep_bend_loss(
    rmin: float = 1.0, rmax: float = 5, steps: int = 3, index: int = 0, **kwargs
) -> Tuple[np.ndarray, np.ndarray]:
    """Returns overlap integral loss.

    The loss is squared because you hit the bend loss twice
    (from bend to straight and from straight to bend)

    FIXME! fix overlap integral code.

    Args:
        rmin: min bend radius (um).
        rmax: max bend radius (um).
        steps: number of steps.
        index: mode index.

    Keyword Args:
        wavelength: (um).
        wg_width: waveguide width.
        t_wg: thickness waveguide (um).
        t_slab: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        w_sim: width simulation (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
    """

    r = np.linspace(rmin, rmax, steps)
    integral = np.zeros_like(r)

    strip = Waveguide(**kwargs)
    strip.compute_modes()

    for i, radius in enumerate(r):
        strip_bend = Waveguide(bend_radius=radius, **kwargs)
        strip_bend.compute_modes()

        mode1_Ey = strip.Ey[index]
        mode1_Ex = strip.Ex[index]
        mode1_Hx = strip.Hx[index]
        mode1_Hy = strip.Hy[index]

        mode2_Ey = strip_bend.Ey[index]
        mode2_Ex = strip_bend.Ex[index]
        mode2_Hx = strip_bend.Hx[index]
        mode2_Hy = strip_bend.Hy[index]

        integrand = (
            np.conj(mode1_Ex) * mode2_Hy
            - np.conj(mode1_Ey) * mode2_Hx
            + mode2_Ex * np.conj(mode1_Hy)
            - mode2_Ey * np.conj(mode1_Hx)
        )

        # square because you hit the bend loss twice
        integral[i] = np.trapz(np.trapz(integrand, axis=0), axis=0) ** 2
    return r, integral


def find_modes(
    wavelength: float = 1.55,
    wg_width: float = 0.45,
    t_wg: float = 0.22,
    t_slab: float = 0.0,
    t_box: float = 2.0,
    t_clad: float = 2.0,
    ncore: Union[float, Callable[[str], float]] = si,
    nclad: Union[float, Callable[[str], float]] = sio2,
    w_sim: float = 2.0,
    resolution: int = 100,
    nmodes: int = 4,
    bend_radius: Optional[float] = None,
    cache: Optional[PathType] = CONFIG["modes"],
) -> Waveguide:
    """
    Args:
        wavelength: (um).
        wg_width: waveguide width.
        t_wg: thickness waveguide (um).
        t_slab: thickness slab (um).
        t_box: thickness BOX (um).
        t_clad: thickness cladding (um).
        ncore: core refractive index.
        nclad: cladding refractive index.
        w_sim: width simulation (um).
        resolution: pixels/um.
        nmodes: number of modes to compute.
        bend_radius: optional bend radius (um).
        cache: directory path to cache modes. None disables the file cache.

    """
    settings = dict(
        wavelength=wavelength,
        wg_width=wg_width,
        t_wg=t_wg,
        t_slab=t_slab,
        t_box=t_box,
        t_clad=t_clad,
        ncore=ncore,
        nclad=nclad,
        w_sim=w_sim,
        resolution=resolution,
        nmodes=nmodes,
        bend_radius=bend_radius,
    )

    if cache:
        cache = pathlib.Path(cache)
        cache.mkdir(exist_ok=True, parents=True)
        filepath = cache / f"{get_hash(settings)}.pkl"
        if filepath.exists():
            logger.info(f"load {filepath} waveguide from file cache.")
            return pickle_load(filepath)
        else:
            waveguide = Waveguide(**settings)
            waveguide.compute_modes()
            waveguide.pickle_dump(filepath=filepath)

    waveguide = Waveguide(**settings)
    waveguide.compute_modes()
    return waveguide


if __name__ == "__main__":
    # c = Waveguide(t_slab=0, ncore=2.00, nclad=1.44, w_sim=5)
    # c.plot_index()

    nitride = find_modes(wavelength=1.55, wg_width=1.0, t_wg=0.4, ncore=2.0, w_sim=5)
    nitride.plot_index()

    # c = pickle_load("strip.pkl")

    # c0 = Waveguide(t_slab=0)
    # c0.plot_Ex(index=0)
    # c0.pickle_dump("strip.pkl")

    # c1 = Waveguide(t_slab=0, bend_radius=5)
    # c1.plot_Ex()
    # c1.pickle_dump("strip_bend5.pkl")

    # c = Waveguide(t_slab=90e-3, bend_radius=5)
    # c.plot_index()

    # r, integral = sweep_bend_loss()
    # plt.plot(r, integral / max(integral), ".")
    # plt.xlabel("bend radius (um)")
    # plt.show()
